--[[--------------------------------------------------
SideBar.lua
Authors: Frank Wunderlich, mozers, VladVRO, frs, BioInfo, Tymur Gubayev
version 1.13
------------------------------------------------------
  Note: Require gui.dll <http://scite-ru.googlecode.com/svn/trunk/lualib/gui/>
               lpeg.dll <http://scite-ru.googlecode.com/svn/trunk/lualib/lpeg/>
              shell.dll <http://scite-ru.googlecode.com/svn/trunk/lualib/shell/>
             COMMON.lua (function GetCurrentWord)
  Connection:
   In file SciTEStartup.lua add a line:
      dofile (props["SciteDefaultHome"].."\\tools\\SideBar.lua")
   Set in a file .properties:
      command.checked.17.*=$(sidebar.show)
      command.name.17.*=SideBar
      command.17.*=SideBar_ShowHide
      command.mode.17.*=subsystem:lua,savebefore:no

    # Set show(1) or hide(0) to SciTE start
    sidebar.show=1
    # Set default settings for Functions/Procedures List
    sidebar.functions.flags=1
    sidebar.functions.params=1
--]]--------------------------------------------------
--[[
Changes by me (leledumbo):
- September 11th 2009
  1. Drop lpeg dependency, use SciTE built-in regular expressions (POSIX)
     to parse source using lexer specific regex (props['routines.regex.<lexer>'])
- November 15th 2009
  2. Move the least frequently used tab FileMan as last tab
  3. Enclose routines.regex.<lexer> between '^\s*' and '.*$' so editor:match will
     match line by line to avoid junk got inserted into routines tab. Consequences:
     one line function will be added as a whole instead of only its header
- November 21th 2009
  4. Use rex_pcre as regular expression parser due to limitation of built-in regex
     and lua patterns
  5. Fix double clicking on Routines tab entry to scroll the lines to centerize the
     caret vertically (not fully fixed)
]]
require 'gui'
require 'shell'

-- you can choose to make it a stand-alone window; just uncomment this line:
-- local win = true

-- local _DEBUG = true --   
--  /  :
local _show_flags = tonumber(props['sidebar.functions.flags']) == 1
local _show_params = tonumber(props['sidebar.functions.params']) == 1

local tab_index = 0
local panel_width = 200
local win_height = props['position.height']
if win_height == '' then win_height = 600 end

local style = props['style.*.32']
local colorback = style:match('back:(#%x%x%x%x%x%x)')
local colorfore
if colorback then
  colorfore = style:match('fore:(#%x%x%x%x%x%x)')
  if colorfore == nil then colorfore = '' end
end

----------------------------------------------------------
-- Common functions
----------------------------------------------------------
local function ReplaceWithoutCase(text, s_find, s_rep)
  local i, j = 1
  local replaced = nil
  repeat
    i, j = text:lower():find(s_find:lower(), j, true)
    if j == nil then return text, replaced end
    text = text:sub(1, i-1)..s_rep..text:sub(j+1)
    replaced = true
  until false
end

local function ShowCompactedLine(line_num)
  local function GetFoldLine(ln)
    while editor.FoldExpanded[ln] do ln = ln-1 end
    return ln
  end
  while not editor.LineVisible[line_num] do
    local x = GetFoldLine(line_num)
    editor:ToggleFold(x)
    line_num = x - 1
  end
end

if _DEBUG then
local nametotime = {} -- maps names to starttimes
  _DEBUG = {}

  _DEBUG.timerstart = function (name)
    nametotime[name] = os.clock()
  end -- _DEBUG.timerstart

  _DEBUG.timer = function (name,...)
    if nametotime[name] then
      local d = os.clock() - nametotime[name]
      print(name,('%.5fs'):format(d),...)
    end
    return d
  end -- _DEBUG.timer

  _DEBUG.timerstop = function (name,...)
    local d = _DEBUG.timer(name,...)
    nametotime[name] = nil
    return d
  end --_DEBUG.timerstop

else
  _DEBUG = {}
  local empty = function (...) end
  _DEBUG.timer, _DEBUG.timerstart, _DEBUG.timerstop = empty, empty, empty
end

----------------------------------------------------------
-- Create panels
----------------------------------------------------------
local tab0 = gui.panel(panel_width + 18)

local memo_path = gui.memo()
tab0:add(memo_path, "top", 22)
if colorback then memo_path:set_memo_colour('', colorback) end

local list_dir_height = win_height/3
if list_dir_height <= 0 then list_dir_height = 200 end

local list_dir = gui.list()
tab0:client(list_dir)
if colorback then list_dir:set_list_colour(colorfore,colorback) end

tab0:context_menu {
  'FileMan: Change Dir|FileMan_ChangeDir',
  'FileMan: Show All|FileMan_MaskAllFiles',
  'FileMan: Only current ext|FileMan_MaskOnlyCurrentExt',
  '', -- separator
  'FileMan: Copy to...|FileMan_FileCopy',
  'FileMan: Move to...|FileMan_FileMove',
  'FileMan: Rename|FileMan_FileRename',
  'FileMan: Delete\tDel|FileMan_FileDelete',
  'FileMan: Execute|FileMan_FileExec',
  'FileMan: Exec with Params|FileMan_FileExecWithParams'
}
-------------------------
local tab1 = gui.panel(panel_width + 18)

local list_func_height = win_height/3
if list_func_height <= 0 then list_func_height = 200 end

local list_func = gui.list(true)
list_func:add_column("Line", 40)
list_func:add_column("Functions/Procedures", 600)
tab1:client(list_func)
if colorback then list_func:set_list_colour(colorfore,colorback) end

tab1:context_menu {
  'Functions: Sort by Order|Functions_SortByOrder',
  'Functions: Sort by Name|Functions_SortByName',
  'Functions: Show/Hide Flags|Functions_ToggleFlags',
  'Functions: Show/Hide Parameters|Functions_ToggleParams',
}
-------------------------
local tab2 = gui.panel(panel_width + 18)

local list_abbrev = gui.list(true)
list_abbrev:add_column("Abbrev", 60)
list_abbrev:add_column("Expansion", 600)
tab2:client(list_abbrev)
if colorback then list_abbrev:set_list_colour(colorfore,colorback) end

-------------------------
local win_parent
if win then
  win_parent = gui.window "Side Bar"
else
  win_parent = gui.panel(panel_width)
end

local tabs = gui.tabbar(win_parent)
tabs:add_tab("Routines", tab1)
tabs:add_tab("Abbrev", tab2)
tabs:add_tab("FileMan", tab0)
win_parent:client(tab0)
win_parent:client(tab2)
win_parent:client(tab1)

if tonumber(props['sidebar.show'])==1 then
  if win then
    win_parent:size(panel_width + 24, 600)
    win_parent:show()
  else
    gui.set_panel(win_parent,"left")
  end
end

----------------------------------------------------------
-- tab0:memo_path   Path and Mask
----------------------------------------------------------
local current_path = ''
local file_mask = '*.*'

local function FileMan_ShowPath()
  local rtf = [[{\rtf\ansi\ansicpg1251{\fonttbl{\f0\fcharset204 Helv;}}{\colortbl;\red0\green0\blue255;\red255\green0\blue0;}\f0\fs16]]
  local path = '\\cf1'..current_path:gsub('\\', '\\\\')
  local mask = '\\cf2'..file_mask..'}'
  memo_path:set_text(rtf..path..mask)
end

----------------------------------------------------------
-- tab0:list_dir   File Manager
----------------------------------------------------------
local function FileMan_ListFILL()
  if current_path == '' then return end
  local folders = gui.files(current_path..'*', true)
  if not folders then return end
  list_dir:clear()
  list_dir:add_item ('[..]', {'..','d'})
  for i, d in ipairs(folders) do
    list_dir:add_item('['..d..']', {d,'d'})
  end
  local files = gui.files(current_path..file_mask)
  if files then
    for i, filename in ipairs(files) do
      list_dir:add_item(filename, {filename})
    end
  end
  list_dir:set_selected_item(0)
  FileMan_ShowPath()
end

local function FileMan_GetSelectedItem()
  local idx = list_dir:get_selected_item()
  if idx == -1 then return '' end
  local data = list_dir:get_item_data(idx)
  local dir_or_file = data[1]
  local attr = data[2]
  return dir_or_file, attr
end

function FileMan_ChangeDir()
  local newPath = gui.select_dir_dlg('Please change current directory', current_path)
  if newPath == nil then return end
  if newPath:match('[\\/]$') then
    current_path = newPath
  else
    current_path = newPath..'\\'
  end
  FileMan_ListFILL()
end

function FileMan_MaskAllFiles()
  file_mask = '*.*'
  FileMan_ListFILL()
end

function FileMan_MaskOnlyCurrentExt()
  local filename, attr = FileMan_GetSelectedItem()
  if filename == '' then return end
  if attr == 'd' then return end
  file_mask = '*.'..filename:gsub('.+%.','')
  FileMan_ListFILL()
end

function FileMan_FileCopy()
  local filename = FileMan_GetSelectedItem()
  if filename == '' or filename == '..' then return end
  local path_destination = gui.select_dir_dlg("Copy to...")
  if path_destination == nil then return end
  os_copy(current_path..filename, path_destination..'\\'..filename)
  FileMan_ListFILL()
end

function FileMan_FileMove()
  local filename = FileMan_GetSelectedItem()
  if filename == '' or filename == '..' then return end
  local path_destination = gui.select_dir_dlg("Move to...")
  if path_destination == nil then return end
  os.rename(current_path..filename, path_destination..'\\'..filename)
  FileMan_ListFILL()
end

function FileMan_FileRename()
  local filename = FileMan_GetSelectedItem()
  if filename == '' or filename == '..' then return end
  local filename_new = shell.inputbox("Rename", "Enter new file name:", filename, function(name) return not name:match('[\\/:|*?"<>]') end)
  if filename_new == nil then return end
  if filename_new ~= '' and filename_new ~= filename then
    os.rename(current_path..filename, current_path..filename_new)
    FileMan_ListFILL()
  end
end

function FileMan_FileDelete()
  local filename, attr = FileMan_GetSelectedItem()
  if filename == '' then return end
  if attr == 'd' then return end
  if shell.msgbox("Are you sure you want to DELETE this file?\n"..filename, "DELETE", 4+256) == 6 then
  -- if gui.message("Are you sure you want to DELETE this file?\n"..filename, "query") then
    os.remove(current_path..filename)
    FileMan_ListFILL()
  end
end

local function FileMan_FileExecWithSciTE(cmd, mode)
  local p0 = props["command.0.*"]
  local p1 = props["command.mode.0.*"]
  props["command.name.0.*"] = 'tmp'
  props["command.0.*"] = cmd
  if mode == nil then mode = 'console' end
  props["command.mode.0.*"] = 'subsystem:'..mode..',savebefore:no'
  scite.MenuCommand(9000)
  props["command.0.*"] = p0
  props["command.mode.0.*"] = p1
end

function FileMan_FileExec(params)
  if params == nil then params = '' end
  local filename = FileMan_GetSelectedItem()
  if filename == '' then return end
  local file_ext = filename:match("[^.]+$")
  if file_ext == nil then return end
  file_ext = '%*%.'..string.lower(file_ext)
  local cmd = ''
  local function CommandBuild(lng)
    local cmd = props['command.build.$(file.patterns.'..lng..')']
    cmd = cmd:gsub(props["FilePath"], current_path..filename)
    return cmd
  end
  -- Lua
  if string.match(props['file.patterns.lua'], file_ext) ~= nil then
    dostring(params)
    dofile(current_path..filename)
  -- Batch
  elseif string.match(props['file.patterns.batch'], file_ext) ~= nil then
    FileMan_FileExecWithSciTE(CommandBuild('batch'))
    return
  -- WSH
  elseif string.match(props['file.patterns.wscript']..props['file.patterns.wsh'], file_ext) ~= nil then
    FileMan_FileExecWithSciTE(CommandBuild('wscript'))
  -- Other
  else
    local ret, descr = shell.exec(current_path..filename..params)
    if not ret then
      print (">Exec: "..filename)
      print ("Error: "..descr)
    end
  end
end

function FileMan_FileExecWithParams()
  if scite.ShowParametersDialog('Exec "'..FileMan_GetSelectedItem()..'". Please set params:') then
    local params = ''
    for p = 1, 4 do
      local ps = props[tostring(p)]
      if ps ~= '' then
        params = params..' '..ps
      end
    end
    FileMan_FileExec(params)
  end
end

local function OpenFile(filename)
  if filename:match(".session$") ~= nil then
    filename = filename:gsub('\\','\\\\')
    scite.Perform ("loadsession:"..filename)
  else
    scite.Open(filename)
  end
  gui.pass_focus()
end

local function FileMan_OpenItem()
  local dir_or_file, attr = FileMan_GetSelectedItem()
  if dir_or_file == '' then return end
  if attr == 'd' then
    gui.chdir(dir_or_file)
    if dir_or_file == '..' then
      local new_path = current_path:gsub('(.*\\).*\\$', '%1')
      if not gui.files(new_path..'*',true) then return end
      current_path = new_path
    else
      current_path = current_path..dir_or_file..'\\'
    end
    FileMan_ListFILL()
  else
    OpenFile(current_path..dir_or_file)
  end
end

list_dir:on_double_click(function()
  FileMan_OpenItem()
end)

list_dir:on_key(function(key)
  if key == 13 then -- Enter
    FileMan_OpenItem()
  elseif key == 8 then -- BackSpace
    list_dir:set_selected_item(0)
    FileMan_OpenItem()
  elseif key == 46 then -- Delete
    FileMan_FileDelete()
  elseif key == 45 then -- Insert
    Favorites_AddFile()
  end
end)

----------------------------------------------------------
-- tab1:list_func   Functions/Procedures
----------------------------------------------------------
local table_functions = {}
-- 1 - function names
-- 2 - line number
-- 3 - function paramaters with parentheses
local _sort = 'order'
local _backjumppos -- store position if jumping

local function Functions_GetNames()
  _DEBUG.timerstart('Functions_GetNames')
  table_functions = {}
  if editor.Length == 0 then return end
  local RoutinesRegex = props['routines.regex.'..props['FileExt']]
  if RoutinesRegex ~= '' then
    RoutinesRegex = '('..RoutinesRegex..')'
    local rex = require 'rex_pcre'
    local IgnoreCase = 0
    if props['routines.regex.'..props['FileExt']..'.ignorecase'] then
      IgnoreCase = PCRE_CASELESS
    end
    local i = 1
    for func in rex.gmatch(editor:GetText(),RoutinesRegex,IgnoreCase) do
      table.insert(table_functions,{func,i})
      i = i + 1
    end
    -- for i=0,LineCount-2 do
      -- local line = editor:GetLine(i)
      -- local func = rex.match(line,RoutinesRegex,1,IgnoreCase)
      -- if func then table.insert(table_functions,{func,i+1}) end
    -- end
  end
  _DEBUG.timerstop('Functions_GetNames','lpeg')
end

local function Functions_ListFILL()
  if tonumber(props['sidebar.show'])~=1 or tab_index~=0 then return end
  if _sort == 'order' then
    table.sort(table_functions, function(a, b) return a[2] < b[2] end)
  else
    table.sort(table_functions, function(a, b) return a[1]:lower() < b[1]:lower() end)
  end
  -- remove duplicates
  for i = #table_functions, 2, -1 do
    if table_functions[i][2] == table_functions[i-1][2] then
      table.remove (table_functions, i)
    end
  end
  list_func:clear()

  local function emptystr(...) return '' end
  local function GetParams (funcitem)
    return (funcitem[3] and ' '..funcitem[3]) or ''
  end
  local function GetFlags (funcitem)
    local res = ''
    local add = ''
    for flag,value in pairs(funcitem) do
      if type(flag)=='string' then
        if type(value)=='string' then  add = flag .. value
        elseif type(value)=='number' then add = flag..':'..value
        else add = flag end
        res = res .. '['.. add ..']'
      end
    end
    if res~='' then res = res .. ' ' end
    return res or ''
  end
  if not _show_params then GetParams = emptystr end
  if not _show_flags then GetFlags = emptystr end

  local function fixname (funcitem)
    return GetFlags(funcitem)..funcitem[1]..GetParams(funcitem)
  end
  for _, a in ipairs(table_functions) do
    list_func:add_item({a[2],fixname(a)}, a[2])
  end
end

function Functions_SortByOrder()
  _sort = 'order'
  Functions_ListFILL()
end

function Functions_SortByName()
  _sort = 'name'
  Functions_ListFILL()
end

function Functions_ToggleParams ()
  _show_params = not _show_params
  Functions_ListFILL()
end

function Functions_ToggleFlags ()
  _show_flags = not _show_flags
  Functions_ListFILL()
end

local function Functions_GotoLine()
  local sel_item = list_func:get_selected_item()
  if sel_item == -1 then return end
  local pos = list_func:get_item_data(sel_item)
  if pos then
    ShowCompactedLine(pos-1)
    local prevpos = editor:LineFromPosition(editor.CurrentPos)
    editor:GotoLine(pos-1)
    local los = nil
    if prevpos + editor.LinesOnScreen / 2 < pos then
      los = editor.LinesOnScreen
    elseif prevpos > pos + editor.LinesOnScreen / 2 then
      los = -editor.LinesOnScreen
    end
    if los then editor:LineScroll(0,los / 2) end
    gui.pass_focus()
  end
end

list_func:on_double_click(function()
  Functions_GotoLine()
end)

list_func:on_key(function(key)
  if key == 13 then -- Enter
    Functions_GotoLine()
  end
end)

----------------------------------------------------------
-- tab2:list_abbrev   Abbreviations
----------------------------------------------------------
local function Abbreviations_ListFILL()
  local function ReadAbbrev(file)
    local abbrev_file = io.open(file)
    if abbrev_file then
      for line in abbrev_file:lines() do
        if line ~= '' then
          local _abr, _exp = line:match('^([^#].-)=(.+)')
          if _abr ~= nil then
            list_abbrev:add_item({_abr, _exp}, _exp)
          else
            local import_file = line:match('^import%s+(.+)')
            if import_file ~= nil then
              ReadAbbrev(file:match('.+\\')..import_file)
            end
          end
        end
      end
      abbrev_file:close()
    end
  end

  list_abbrev:clear()
  local abbrev_file = 'abbreviations.*.'..props['FileExt']
  local abbrev_filename = props[abbrev_file]
  ReadAbbrev(abbrev_filename)
end

local function Abbreviations_InsertExpansion()
  local sel_item = list_abbrev:get_selected_item()
  if sel_item == -1 then return end
  local expansion = list_abbrev:get_item_data(sel_item)
  scite.InsertAbbreviation(expansion)
  gui.pass_focus()
  editor:CallTipCancel()
end

local function Abbreviations_ShowExpansion()
  local sel_item = list_abbrev:get_selected_item()
  if sel_item == -1 then return end
  local expansion = list_abbrev:get_item_data(sel_item)
  expansion = expansion:gsub('\\\\','\4'):gsub('\\r','\r'):gsub('(\\n','\n'):gsub('\\t','\t'):gsub('\4','\\')
  editor:CallTipCancel()
  editor:CallTipShow(editor.CurrentPos, expansion)
end

list_abbrev:on_double_click(function()
  Abbreviations_InsertExpansion()
end)

list_abbrev:on_select(function()
  Abbreviations_ShowExpansion()
end)

list_abbrev:on_key(function(key)
  if key == 13 then -- Enter
    Abbreviations_InsertExpansion()
  end
end)

----------------------------------------------------------
-- Events
----------------------------------------------------------
local line_count

local function OnSwitch()
  _DEBUG.timerstart('OnSwitch')
  line_count = editor.LineCount
  if tab0:bounds() then -- visible FileMan
    local path = props['FileDir']
    if path == '' then return end
    path = path:gsub('\\$','')..'\\'
    if path ~= current_path then
      current_path = path
      FileMan_ListFILL()
    end
  elseif tab1:bounds() then -- visible Funk/Bmk
    Functions_GetNames()
    Functions_ListFILL()
--~     Bookmarks_ListFILL()
  elseif tab2:bounds() then -- visible Abbrev
    Abbreviations_ListFILL()
  end
  _DEBUG.timerstop('OnSwitch')
end

tabs:on_select(function(ind)
  tab_index=ind
  OnSwitch()
end)

--  /  
function SideBar_ShowHide()
  if tonumber(props['sidebar.show'])==1 then
    if win then
      win_parent:hide()
    else
      gui.set_panel()
    end
    props['sidebar.show']=0
  else
    if win then
      win_parent:show()
    else
      gui.set_panel(win_parent,"set_panel")
    end
    props['sidebar.show']=1
    OnSwitch()
  end
end

local function OnDocumentCountLinesChanged(def_line_count)
  if tab1:bounds() then -- visible Funk/Bmk
    local cur_line = editor:LineFromPosition(editor.CurrentPos)
    for i = 1, #table_functions do
      local table_line = table_functions[i][2]
      if table_line > cur_line then
        table_functions[i][2] = table_line + def_line_count
      end
    end
    Functions_ListFILL()

--~     Bookmarks_RefreshTable()
  end
end

-- Add user event handler OnSwitchFile
function SideBar_OnSwitchFile(file)
  local result
  OnSwitch()
  return result
end
scite_OnSwitchFile(SideBar_OnSwitchFile)

-- Add user event handler OnOpen
function SideBar_OnOpen(file)
  local result
  OnSwitch()
  return result
end
scite_OnOpen(SideBar_OnOpen)

-- Add user event handler OnUpdateUI
function SideBar_OnOpenOnUpdateUI()
  local result
  if (editor.Focus and line_count) then
    local line_count_new = editor.LineCount
    local def_line_count = line_count_new - line_count
    if def_line_count ~= 0 then
      OnDocumentCountLinesChanged(def_line_count)
      line_count = line_count_new
    end
  end
  return result
end
scite_OnUpdateUI(SideBar_OnOpenOnUpdateUI)

-- Add user event handler OnSave
local old_OnSave = OnSave
function SideBar_OnOpenOnSave(file)
  local result
  Functions_GetNames()
  Functions_ListFILL()
  return result
end
scite_OnSave(SideBar_OnOpenOnSave)

-- Add user event handler OnSendEditor
local old_OnSendEditor = OnSendEditor
function OnSendEditor(id_msg, wp, lp)
  local result
  if old_OnSendEditor then result = old_OnSendEditor(id_msg, wp, lp) end
  if id_msg == SCI_MARKERADD then
  --[[
    if lp == 1 then Bookmark_Add(wp) Bookmarks_ListFILL() end
  elseif id_msg == SCI_MARKERDELETE then
    if lp == 1 then Bookmark_Delete(wp) Bookmarks_ListFILL() end
  elseif id_msg == SCI_MARKERDELETEALL then
    if wp == 1 then Bookmark_Delete() Bookmarks_ListFILL() end
  ]]
  end
  return result
end

-- Add user event handler OnFinalise
local old_OnFinalise = OnFinalise
function OnFinalise()
  local result
  if old_OnFinalise then result = old_OnFinalise() end
--~   Favorites_SaveList()
  return result
end

----------------------------------------------------------
-- Go to function definition
----------------------------------------------------------

--         (   table_functions)
local function Func2Line(funcname)
  if not next(table_functions) then
    Functions_GetNames()
  end
  for i = 1, #table_functions do
    if funcname == table_functions[i][1] then
      return table_functions[i][2]
    end
  end
end

--      
local function JumpToFuncDefinition()
  local funcname = GetCurrentWord()
  local line = Func2Line(funcname)
  if line then
    _backjumppos = editor.CurrentPos
    editor:GotoLine(line)
    return true
  end
  return false
end

local function JumpBack()
  if not _backjumppos then return false end
  editor:GotoPos(_backjumppos)
  _backjumppos = nil
  return true
end

-- Add user event handler OnDoubleClick
local old_OnDoubleClick = OnDoubleClick
function OnDoubleClick(shift, ctrl, alt)
  local result
  if old_OnDoubleClick then result = old_OnDoubleClick(shift, ctrl, alt) end
  if shift then
    if JumpToFuncDefinition() then return true end
  end
  return result
end

-- Add user event handler OnKey
local old_OnKey = OnKey
function OnKey(key, shift, ctrl, alt, char)
  local result
  if old_OnKey then result = old_OnKey(key, shift, ctrl, alt, char) end
  if (editor.Focus) then
    if ctrl and key == 188 and JumpBack() then return true end --char == ','
    if ctrl and key == 190 and JumpToFuncDefinition() then return true end --char == '.'
  end
  return result
end
